""" Module that contains a main function that checks whether a directed graph
is cyclic or not

Calls to advice insertions are included at join-points. These calls belong to a
abstract class Advisor, which must be implemented by interested parties """

from .. util.advisor import Advisor
from . directed_graph_core import DirectedGraphCore
from . vertex import Vertex
from typing import Set, MutableMapping


def is_cyclic(directed_graph: DirectedGraphCore, advisor: Advisor):
    """ Function that checks whether a directed graph contains a cycle or not

    Args:
        directed_graph (DirectedGraph): The directed graph
        advisor(Advisor): Object that contains advice which can be inserted at
        join points

    Returns:
        bool: True if the directed graph contains a cycle, otherwise False """

    def _is_cyclic_dfs(directed_graph: DirectedGraphCore, vertex: Vertex,
                       visited_already: MutableMapping[Vertex, bool],
                       in_cycle: MutableMapping[Vertex, bool]):
        """ Function that recursively searches the directed graph depth first
        and checks if a vertex was already in_cycle before.

        It checks all vertices that have not been traversed before. The heads
        of those
        vertices are followed. If in that traversal, a vertex is found that is
        present in the dict "in_cycle" with a value of true, then a cycle is
        present

        Args:
            directed_graph (DirectedGraphCore): The directed graph
            vertex(Vertex): The current vertex
            visited_already (dict): A dictionary that maintains whether
                vertices have been traversed already. It's a performance
                measure put in place in order to shortcut processing if a
                vertex was already processed by another subtree in_cycle
                (list): A list that, if a vertex has been found to part be
                part of cycle, for that vertex, has a value of true. If
                that vertex is not part of a cycle, it's value is false

        Returns:
            bool: True if the vertex was in_cycle before, False otherwise """

        visited_already[vertex] = True
        in_cycle[vertex] = True
        advisor.advise("visit_vertex", directed_graph, vertex)

        for edge in vertex.get_edges():
            if visited_already.get(edge.get_head()) is None:
                if _is_cyclic_dfs(directed_graph, edge.get_head(),
                                  visited_already, in_cycle):
                    advisor.advise("cycle_reported_recursive", directed_graph,
                                   edge.get_head())
                    return True
                else:
                    advisor.advise("no_cycle_reported_recursive",
                                   directed_graph, vertex)
            elif in_cycle[edge.get_head()]:
                advisor.advise("cycle_found", directed_graph, vertex,
                               edge.get_head())
                return True
            elif visited_already.get(edge.get_head().get_label()):
                advisor.advise("vertex_already_visited", directed_graph, edge)

        in_cycle[vertex] = False
        return False

    def _is_cyclic_inner(directed_graph: DirectedGraphCore):
        """ Main function that loops through the vertices and starts
        the traversal for each vertex.

        Args:
            directed_graph (DirectedGraphCore): The directed graph """

        visited_already: MutableMapping[Vertex, bool] = dict()
        in_cycle: MutableMapping[Vertex, bool] = {
            vertex: False for vertex in directed_graph.get_vertices()}
        for vertex in directed_graph.get_vertices():
            if vertex not in visited_already:
                if _is_cyclic_dfs(directed_graph, vertex, visited_already,
                                  in_cycle):
                    return True

        return False

    return _is_cyclic_inner(directed_graph)



from __future__ import annotations
from . vertex import Vertex
from copy import deepcopy
from typing import Collection, Set, Mapping, Any, List
from . edge import Edge
from . algorithm_ordering import AlgorithmOrdering

""" Module that contains the definition of a directed graph as a class """


class DirectedGraphCore(object):
    """ Directed graph core class, it just contains the pure defintion of the
    concept, not the extra functionalities, like calculating sccs, etc.. """

    def __init__(self, vertices: Mapping[Any, List[Any]] = None,
                 algorithm_ordering=AlgorithmOrdering.NATURAL):
        """ Initialises a directed graph (with the provided vertices)

        Args:
            vertices(dict): a dict with the vertices and their tails in it
            vertex_sorting: Using the VertexSortOrder definition, algorithms
                access the vertices in the directed graph in a certain order
        """

        self._algorithm_ordering = algorithm_ordering
        self._vertices: Set[Vertex] = set()
        if vertices is not None:
            for label in vertices.keys():
                self.create_add_vertex(label)
            for label, heads in vertices.items():
                for head in heads:
                    self.add_edge(self.get_vertex(label),
                                  self.get_vertex(head))

    def get_vertex(self, label: Any):
        """ Returns the vertex that coincides with the label

        Args:
            label: the label of the vertex

        Returns:
            The vertex object
        """
        for vertex in self._vertices:
            if vertex.get_label() == label:
                return vertex

        raise RuntimeError(f"label {label} couldn't be found in vertices")

    def copy(self) -> DirectedGraphCore:
        """ Copies the directed graph and returns it

        Returns:
            the copied directed graph """

        return deepcopy(self)

    def create_add_vertex(self, label: Any):
        """ Adds a vertex to the dictionary of vertices

        Args:
            label: a vertex represented by its label """

        for vertex in self._vertices:
            if vertex.get_label() == label:
                raise RuntimeError(
                    f"Vertex = {label} is already a vertex in this directed " +
                    " graph")

        self._vertices.add(Vertex(label, self._algorithm_ordering))

    def add_vertex(self, vertex: Vertex):
        """ Function that adds a vertex to the directed graph

        Args:
            vertex: the vertex
        """

        self._vertices.add(vertex)

    def get_vertices(self) -> Collection[Vertex]:
        """ Returns the vertices set

        Args:
            vertex_sorting: Indicates the ordering of vertices that
                will be used in algorithms where appropriate
        Returns
            self._vertices according to the indicated vertex ordering """

        if self._algorithm_ordering == AlgorithmOrdering.NATURAL:
            return self._vertices
        else:
            return sorted(self._vertices,
                          key=lambda vertex: vertex.get_label(),
                          reverse=self._algorithm_ordering ==
                          AlgorithmOrdering.DESC)

    def add_edge(self, tail: Vertex, head: Vertex):
        """ Adds an edge to the graph, the edge is identified by a tail and
        a head vertex.

        Args:
            tail: the edge that represents the start vertex
            head: the edge that represents the destination vertex """

        tail.add_edge(head)
        head.increase_indegree()

    def get_all_edges(self) -> Set[Edge]:
        """ Method that retrieves all edges of all vertices

        Returns:
            set(): A set of all edges in the directed graph """

        return {e for v in self._vertices for e in v.get_edges()}

    def reversed(self, inplace: bool = True) -> DirectedGraphCore:
        """ Function that calculates the transposed graph

        Args:
            inplace: if true, then changes the object it self and
                returns it, if false, returns a copy

        Returns:
            DirectedGraph: The reversed graph """

        graph: DirectedGraphCore = self
        if not inplace:
            graph = self.copy()

        edges: List[Edge] = list()
        for vertex in graph._vertices:
            edges.extend(vertex.get_edges())
            vertex.remove_edges()

        for edge in edges:
            edge.reverse()
            edge.get_tail().add_edge(edge.get_head())

        return graph

    def get_vertices_count(self) -> int:
        return len(self._vertices)

    def __str__(self):
        res = ""
        for vertex in self._vertices:
            res += "\n" + str(vertex)

        return res



""" Module that contains a main function that trails the directed graph.

A directed trail is a walk where edges are not repeated. Edges are traversed
in a certain order, which is dictated by the data structure that holds the
edges.

Calls to advice insertions are included at join-points. These calls belong to a
abstract class Advisor, which must be implemented by interested parties

"""

from .. util.advisor import Advisor
from . directed_graph_core import DirectedGraphCore
from . vertex import Vertex


def trail(directed_graph: DirectedGraphCore, advisor: Advisor):
    """ Main function that walks the directed graph, restricted by the trail
     feature (no edge repetitions)

    Args:
        directed_graph (DirectedGraph): The directed graph
        advisor (Advisor): Object that contains advice which can be inserted
        at join points """

    VISITED = "visited"

    def is_edge_visited(edge):
        if edge.get_attr(VISITED):
            return True
        else:
            return False

    def _trail_dfs(vertex: Vertex):
        """ Function that recursively walks the directed graph

        Args:
            directed_graph (DirectedGraph): The directed graph
            vertex (Vertex): The current vertex
        """

        advisor.advise("visit_vertex", directed_graph, vertex)
        for edge in vertex.get_edges():
            if not is_edge_visited(edge):
                edge.set_attr(VISITED, True)
                advisor.advise("edge_not_visited", directed_graph, edge)
                _trail_dfs(edge.get_head())
            else:
                advisor.advise("edge_visited_already", directed_graph, edge)

        return False

    for vertex in directed_graph.get_vertices():
        _trail_dfs(vertex)



""" Module thatAlgorithmOrderfor the main directed graph functionality
"""

from pythonalgos.graph.algorithm_ordering import AlgorithmOrdering
import unittest
from pythonalgos.graph.directed_graph import DirectedGraph


class TestEdge(unittest.TestCase):

    def setUp(self):
        self.vertices = {0: [1], 1: [2], 2: [], 3: [6, 5, 4], 6: [],
                         5: [], 4: []}
        self.directed_graph = DirectedGraph(self.vertices)

    def test_reverse_edge(self):
        vertex1 = self.directed_graph.get_vertex(1)
        vertex2 = self.directed_graph.get_vertex(2)
        edge = next(iter(vertex1.get_edges()))
        edge.reverse()
        self.assertTrue(edge.get_tail() == vertex2)
        self.assertTrue(edge.get_head() == vertex1)

    def test_get_edges_ascending_order(self):
        self.directed_graph = DirectedGraph(
            self.vertices, algorithm_ordering=AlgorithmOrdering.ASC)
        self.assertTrue(next(iter(
            self.directed_graph.get_vertex(3).get_edges()))
                .get_head().get_label() == 4)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



""" Module that contains test for the trail functionality of the directed graph
"""

import unittest
from pythonalgos.graph.directed_graph import DirectedGraph
import os
import time
from pythonalgos.util.logging import Logging
from pythonalgos.util import path_tools as pt
from pythonalgos.util.advisor import Advisor
from os import path

COUNT = "count"


class TestDirectedGraphTrail(unittest.TestCase):

    def setUp(self):
        pass

    def test_trail(self):
        self.vertices = {0: [1, 2], 1: [3, 4], 2: [5, 6, 10], 3: [], 4: [],
                         5: [], 6: [7, 8], 8: [], 7: [9], 9: [2], 10: [11],
                         11: []}
        self.directed_graph = DirectedGraph(self.vertices)
        self.directed_graph.trail(TestAdvisor())
        for edge in self.directed_graph.get_all_edges():
            self.assertEqual(edge.get_attr(COUNT), 1)

    def tearDown(self):
        pass


class TestAdvisor(Advisor):

    def __init__(self):
        super().__init__()

    def visit_vertex(self, directed_graph, vertex):
        print("visited vertex: {0}".format(vertex.get_label()))

    def edge_not_visited(self, directed_graph, edge):
        count = edge.get_attr(COUNT) or 0
        edge.set_attr(COUNT, count + 1)
        print("edge {0}, {1} not visisted".format(edge.get_tail().get_label(),
              edge.get_head().get_label()))

    def edge_visited_already(self, directed_graph, edge):
        print("edge {0}, {1} already visisted".format(edge.get_tail()
              .get_label(), edge.get_head().get_label()))


if __name__ == '__main__':
    unittest.main()



from typing import Mapping, Any

""" Module that contains the definition of the class that represents an edge in
a directed graph
"""


class Edge(object):
    """ Class that represents an edge in a directed graph. An edge is
    characterized by a tail, a head and a set of attributes.

    The attributes can be anything that the implementation is interested in.
    For example, a weighted graph would have a necessity to put a weight to
    each edge."""

    def __init__(self, tail, head, **attrs: Mapping[str, str]):
        """ Initialises the edge.

        Args:
            tail(Vertex): the tail vertex
            head(Vertex): the head vertex
            **attrs: additional atttributes that define the edge
        """

        self._label = None
        self._tail = tail
        self._head = head
        self._attrs = attrs

    def get_tail(self):
        return self._tail

    def get_head(self):
        return self._head

    def set_tail(self, tail):
        self.tail = tail

    def set_head(self, head):
        self.head = head

    def set_attr(self, attr: str, value: Any):
        self._attrs[attr] = value

    def reverse(self):
        self._tail, self._head = self._head, self._tail

    def get_attr(self, attr: str) -> Any:
        return self._attrs.get(attr)



#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md") as f:
    long_desc = f.read()

setup(name="python-algos",
      version=0.16,
      description="A set of bare algorithms",
      long_description=long_desc,
      long_description_content_type="text/markdown", 
      keywords="algorithms graphs directed graphs cyclic kosaraju tarjan",
      author="Evo Williamson",
      author_email="evowilliamson@gmail.com",
      license="GNU General Public License, version 2",
      url="https://github.com/evowilliamson/python-algos",
      packages=find_packages(),
      install_requires=[],
      test_suite="tests",
      classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Topic :: Software Development",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.4",
        "Operating System :: OS Independent"
      ])



from pathlib import Path
import os
import shutil
import sys


def create_dir_in_user_home(dir, overwrite=True):
    if overwrite:
        clean_dir_in_user_home(dir)
    os.mkdir(Path(get_user_home(), dir))


def clean_dir_in_user_home(dir):
    if Path(str(get_user_home()), dir).exists() and \
            Path(str(get_user_home()), dir).is_dir():
        shutil.rmtree(Path(get_user_home(), dir))


def get_dir_in_user_home(dir):
    return str(Path(get_user_home(), dir))


def get_user_home():
    return str(Path.home())


def append_to_path(top_parent_package):
    """ Function that searches up the folder tree and finds the top package and
    then adds the package to sys.path

    Args:
        top_parent_package(str): The top parent package to find

    """

    name = os.path.realpath(__file__)
    while name.split("/")[-1] != top_parent_package:
        name = os.path.dirname(name)
    sys.path.append(os.path.dirname(name))



from __future__ import annotations
from . vertex import Vertex
from . import kosaraju_sccs
from . import cyclic as cyclic
from . import directed_trail as trail
from . directed_graph_core import DirectedGraphCore
from copy import deepcopy
from .. util.advisor import Advisor
from typing import Any, List, Mapping, Set, Collection
from . edge import Edge
from . algorithm_ordering import AlgorithmOrdering


""" Module that contains the definition of a directed graph as a class """


class DirectedGraph(object):
    """ Class to represent directed graphs.
    https://en.wikipedia.org/wiki/Directed_graph """

    def __init__(self, vertices: Mapping[Any, List[Any]] = None,
                 algorithm_ordering=AlgorithmOrdering.NATURAL):
        """ Initialises a directed graph (with the provided vertices)

        Args:
            vertices(dict): a dict with the vertices and their tails in it
        """

        self.directed_graph = DirectedGraphCore(vertices, algorithm_ordering)

    def copy(self) -> DirectedGraph:
        """ Copies the directed graph and returns it

        Returns:
            the copied directed graph """

        return deepcopy(self)

    def get_vertex(self, label: Any):
        """ Returns the vertex that coincides with the label

        Args:
            label: the label of the vertex

        Returns:
            The vertex object
        """

        return self.directed_graph.get_vertex(label)

    def add_vertex(self, label: Any):
        """ Adds a vertex to the dictionary of vertices

        Args:
            label: a vertex represented by its label """

        self.directed_graph.create_add_vertex(label)

    def get_vertices(self) -> Collection[Vertex]:
        """ Returns the vertices dictionary

        Returns
            self._vertices (dict) """

        return self.directed_graph.get_vertices()

    def add_edge(self, tail: Any, head: Any):
        """ Adds an edge to the graph, the edge is identified by a tail and
        a head vertex.

        Args:
            tail: the edge that represents the start vertex
            head: the edge that represents the destination vertex """

        self.directed_graph.add_edge(tail, head)

    def get_all_edges(self) -> Set[Edge]:
        """ Method that retrieves all edges of all vertices

        Returns:
            set(): A set of all edges in the directed graph """

        return self.directed_graph.get_all_edges()

    def get_vertices_count(self) -> int:
        return self.directed_graph.get_vertices_count()

    def __str__(self):
        return self.directed_graph.__str__()

    def create_sccs_kosaraju_dfs(
            self, nontrivial: bool = True) -> List[Set[Vertex]]:
        return kosaraju_sccs.create_sccs_kosaraju_dfs(self.directed_graph,
                                                      nontrivial)

    def is_cyclic(self, advisor: Advisor = Advisor()):
        """ Method that uses a helper module to check for cycles in the
        directed graph.

        Args:
            advisor(Advisor): The class that implements the advice that is to
            be inserted at join points in the algorith. The default advice is
            empty """

        return cyclic.is_cyclic(self.directed_graph, advisor)

    def trail(self, advisor: Advisor = Advisor()):
        """ Method that trails the directed graph

         Args:
            advisor(Advisor): The class that implements the advice that is to
            be inserted at join points in the algorith. The default advice is
            empty """

        return trail.trail(self.directed_graph, advisor)

    def reversed(self, inplace=True) -> DirectedGraphCore:
        return self.directed_graph.reversed(inplace)

    def get_direct_graph_core(self):
        return self.directed_graph
